#include "pls/internal/scheduling/scheduler.h"
#include "pls/internal/scheduling/scheduler_memory.h"

using namespace pls::internal::scheduling;

#include <iostream>
#include <complex>
#include <vector>

#include "benchmark_runner.h"
#include "benchmark_base/fib.h"

using namespace comparison_benchmarks::base;

int pls_fib(int n) {
  if (n <= 1) {
    return 1;
  }

  int a = pls_fib(n - 1);
  int b = pls_fib(n - 2);

  return a + b;
}

constexpr int MAX_NUM_THREADS = 1;
constexpr int MAX_NUM_TASKS = 64;
constexpr int MAX_STACK_SIZE = 128;

int main(int argc, char **argv) {
  int num_threads;
  string directory;
  benchmark_runner::read_args(argc, argv, num_threads, directory);

  string test_name = to_string(num_threads) + ".csv";
  string full_directory = directory + "/PLS_v3/";
  benchmark_runner runner{full_directory, test_name};

  static_scheduler_memory<MAX_NUM_THREADS,
                          MAX_NUM_TASKS,
                          MAX_STACK_SIZE> static_scheduler_memory;

  scheduler scheduler{static_scheduler_memory, (unsigned) num_threads};

  volatile int res;
  for (int i = 0; i < fib::NUM_WARMUP_ITERATIONS; i++) {
    scheduler.perform_work([&]() {
      res = pls_fib(fib::INPUT_N);
    });
  }

  for (int i = 0; i < fib::NUM_ITERATIONS; i++) {
    scheduler.perform_work([&]() {
      runner.start_iteration();
      res = pls_fib(fib::INPUT_N);
      runner.end_iteration();
    });
  }
  runner.commit_results(true);

  return 0;
}
